@@ -1,6 +1,6 @@
 /* connection_int.c - code used by the connection object
  *
- * Copyright (C) 2003-2010 Federico Di Gregorio <fog@debian.org>
+ * Copyright (C) 2003-2019 Federico Di Gregorio <fog@debian.org>
  *
  * This file is part of psycopg.
  *
@@ -79,7 +79,7 @@ const int SRV_STATE_UNCHANGED = -1;
 PyObject *
 conn_text_from_chars(connectionObject *self, const char *str)
 {
-    return psycopg_text_from_chars_safe(str, -1, self ? self->pydecoder : NULL);
+    return psyco_text_from_chars_safe(str, -1, self ? self->pydecoder : NULL);
 }
 
 
@@ -121,10 +121,6 @@ exit:
 PyObject *
 conn_decode(connectionObject *self, const char *str, Py_ssize_t len)
 {
-    PyObject *b = NULL;
-    PyObject *t = NULL;
-    PyObject *rv = NULL;
-
     if (len < 0) { len = strlen(str); }
 
     if (self) {
@@ -132,22 +128,28 @@ conn_decode(connectionObject *self, const char *str, Py_ssize_t len)
             return self->cdecoder(str, len, NULL);
         }
         else if (self->pydecoder) {
-            if (!(b = Bytes_FromStringAndSize(str, len))) { goto exit; }
+            PyObject *b = NULL;
+            PyObject *t = NULL;
+            PyObject *rv = NULL;
+
+            if (!(b = Bytes_FromStringAndSize(str, len))) { goto error; }
             if (!(t = PyObject_CallFunctionObjArgs(self->pydecoder, b, NULL))) {
-                goto exit;
+                goto error;
             }
-            rv = PyTuple_GetItem(t, 0);
-            Py_XINCREF(rv);
+            if (!(rv = PyTuple_GetItem(t, 0))) { goto error; }
+            Py_INCREF(rv);      /* PyTuple_GetItem gives a borrowes one */
+error:
+            Py_XDECREF(t);
+            Py_XDECREF(b);
+            return rv;
+        }
+        else {
+            return PyUnicode_FromStringAndSize(str, len);
         }
     }
     else {
         return PyUnicode_FromStringAndSize(str, len);
     }
-
-exit:
-    Py_XDECREF(t);
-    Py_XDECREF(b);
-    return rv;
 }
 
 /* conn_notice_callback - process notices */
@@ -476,7 +478,7 @@ conn_get_python_codec(const char *encoding,
 
     /* get the Python name of the encoding as a C string */
     if (!(encname = conn_pgenc_to_pyenc(encoding, &pgenc))) { goto exit; }
-    if (!(encname = psycopg_ensure_bytes(encname))) { goto exit; }
+    if (!(encname = psyco_ensure_bytes(encname))) { goto exit; }
 
     /* Look up the codec functions */
     if (!(enc_tmp = PyCodec_Encoder(Bytes_AS_STRING(encname)))) { goto exit; }
@@ -647,25 +649,23 @@ conn_is_datestyle_ok(PGconn *pgconn)
 /* conn_setup - setup and read basic information about the connection */
 
 RAISES_NEG int
-conn_setup(connectionObject *self, PGconn *pgconn)
+conn_setup(connectionObject *self)
 {
-    PGresult *pgres = NULL;
-    char *error = NULL;
     int rv = -1;
 
-    self->equote = conn_get_standard_conforming_strings(pgconn);
-    self->server_version = conn_get_server_version(pgconn);
+    self->equote = conn_get_standard_conforming_strings(self->pgconn);
+    self->server_version = conn_get_server_version(self->pgconn);
     self->protocol = conn_get_protocol_version(self->pgconn);
     if (3 != self->protocol) {
         PyErr_SetString(InterfaceError, "only protocol 3 supported");
         goto exit;
     }
 
-    if (0 > conn_read_encoding(self, pgconn)) {
+    if (0 > conn_read_encoding(self, self->pgconn)) {
         goto exit;
     }
 
-    if (0 > conn_setup_cancel(self, pgconn)) {
+    if (0 > conn_setup_cancel(self, self->pgconn)) {
         goto exit;
     }
 
@@ -676,11 +676,10 @@ conn_setup(connectionObject *self, PGconn *pgconn)
     if (!dsn_has_replication(self->dsn) && !conn_is_datestyle_ok(self->pgconn)) {
         int res;
         Py_UNBLOCK_THREADS;
-        res = pq_set_guc_locked(self, "datestyle", "ISO",
-            &pgres, &error, &_save);
+        res = pq_set_guc_locked(self, "datestyle", "ISO", &_save);
         Py_BLOCK_THREADS;
         if (res < 0) {
-            pq_complete_error(self, &pgres, &error);
+            pq_complete_error(self);
             goto unlock;
         }
     }
@@ -708,7 +707,6 @@ exit:
 static int
 _conn_sync_connect(connectionObject *self)
 {
-    PGconn *pgconn;
     int green;
 
     /* store this value to prevent inconsistencies due to a change
@@ -716,31 +714,31 @@ _conn_sync_connect(connectionObject *self)
     green = psyco_green();
     if (!green) {
         Py_BEGIN_ALLOW_THREADS;
-        self->pgconn = pgconn = PQconnectdb(self->dsn);
+        self->pgconn = PQconnectdb(self->dsn);
         Py_END_ALLOW_THREADS;
-        Dprintf("conn_connect: new postgresql connection at %p", pgconn);
+        Dprintf("conn_connect: new PG connection at %p", self->pgconn);
     }
     else {
         Py_BEGIN_ALLOW_THREADS;
-        self->pgconn = pgconn = PQconnectStart(self->dsn);
+        self->pgconn = PQconnectStart(self->dsn);
         Py_END_ALLOW_THREADS;
-        Dprintf("conn_connect: new green postgresql connection at %p", pgconn);
+        Dprintf("conn_connect: new green PG connection at %p", self->pgconn);
     }
 
-    if (pgconn == NULL)
+    if (!self->pgconn)
     {
         Dprintf("conn_connect: PQconnectdb(%s) FAILED", self->dsn);
         PyErr_SetString(OperationalError, "PQconnectdb() failed");
         return -1;
     }
-    else if (PQstatus(pgconn) == CONNECTION_BAD)
+    else if (PQstatus(self->pgconn) == CONNECTION_BAD)
     {
         Dprintf("conn_connect: PQconnectdb(%s) returned BAD", self->dsn);
-        PyErr_SetString(OperationalError, PQerrorMessage(pgconn));
+        PyErr_SetString(OperationalError, PQerrorMessage(self->pgconn));
         return -1;
     }
 
-    PQsetNoticeProcessor(pgconn, conn_notice_callback, (void*)self);
+    PQsetNoticeProcessor(self->pgconn, conn_notice_callback, (void*)self);
 
     /* if the connection is green, wait to finish connection */
     if (green) {
@@ -757,7 +755,7 @@ _conn_sync_connect(connectionObject *self)
      */
     self->status = CONN_STATUS_READY;
 
-    if (conn_setup(self, self->pgconn) == -1) {
+    if (conn_setup(self) == -1) {
         return -1;
     }
 
@@ -861,11 +859,16 @@ _conn_poll_connecting(connectionObject *self)
 /* Advance to the next state after an attempt of flushing output */
 
 static int
-_conn_poll_advance_write(connectionObject *self, int flush)
+_conn_poll_advance_write(connectionObject *self)
 {
     int res;
+    int flush;
 
     Dprintf("conn_poll: poll writing");
+
+    flush = PQflush(self->pgconn);
+    Dprintf("conn_poll: PQflush() = %i", flush);
+
     switch (flush) {
     case  0:  /* success */
         /* we've finished pushing the query to the server. Let's start
@@ -889,18 +892,24 @@ _conn_poll_advance_write(connectionObject *self, int flush)
     return res;
 }
 
-/* Advance to the next state after a call to a pq_is_busy* function */
+
+/* Advance to the next state after reading results */
+
 static int
-_conn_poll_advance_read(connectionObject *self, int busy)
+_conn_poll_advance_read(connectionObject *self)
 {
     int res;
+    int busy;
 
     Dprintf("conn_poll: poll reading");
+
+    busy = pq_get_result_async(self);
+
     switch (busy) {
     case 0: /* result is ready */
-        res = PSYCO_POLL_OK;
         Dprintf("conn_poll: async_status -> ASYNC_DONE");
         self->async_status = ASYNC_DONE;
+        res = PSYCO_POLL_OK;
         break;
     case 1: /* result not ready: fd would block */
         res = PSYCO_POLL_READ;
@@ -909,13 +918,15 @@ _conn_poll_advance_read(connectionObject *self, int busy)
         res = PSYCO_POLL_ERROR;
         break;
     default:
-        Dprintf("conn_poll: unexpected result from pq_is_busy: %d", busy);
+        Dprintf("conn_poll: unexpected result from pq_get_result_async: %d",
+            busy);
         res = PSYCO_POLL_ERROR;
         break;
     }
     return res;
 }
 
+
 /* Poll the connection for the send query/retrieve result phase
 
   Advance the async_status (usually going WRITE -> READ -> DONE) but don't
@@ -929,27 +940,18 @@ _conn_poll_query(connectionObject *self)
     switch (self->async_status) {
     case ASYNC_WRITE:
         Dprintf("conn_poll: async_status = ASYNC_WRITE");
-        res = _conn_poll_advance_write(self, PQflush(self->pgconn));
+        res = _conn_poll_advance_write(self);
         break;
 
     case ASYNC_READ:
         Dprintf("conn_poll: async_status = ASYNC_READ");
-        if (self->async) {
-            res = _conn_poll_advance_read(self, pq_is_busy(self));
-        }
-        else {
-            /* we are a green connection being polled as result of a query.
-              this means that our caller has the lock and we are being called
-              from the callback. If we tried to acquire the lock now it would
-              be a deadlock. */
-            res = _conn_poll_advance_read(self, pq_is_busy_locked(self));
-        }
+        res = _conn_poll_advance_read(self);
         break;
 
     case ASYNC_DONE:
         Dprintf("conn_poll: async_status = ASYNC_DONE");
         /* We haven't asked anything: just check for notifications. */
-        res = _conn_poll_advance_read(self, pq_is_busy(self));
+        res = _conn_poll_advance_read(self);
         break;
 
     default:
@@ -972,7 +974,6 @@ static int
 _conn_poll_setup_async(connectionObject *self)
 {
     int res = PSYCO_POLL_ERROR;
-    PGresult *pgres;
 
     switch (self->status) {
     case CONN_STATUS_CONNECTING:
@@ -1023,12 +1024,12 @@ _conn_poll_setup_async(connectionObject *self)
         res = _conn_poll_query(self);
         if (res == PSYCO_POLL_OK) {
             res = PSYCO_POLL_ERROR;
-            pgres = pq_get_last_result(self);
-            if (pgres == NULL || PQresultStatus(pgres) != PGRES_COMMAND_OK ) {
+            if (self->pgres == NULL
+                    || PQresultStatus(self->pgres) != PGRES_COMMAND_OK ) {
                 PyErr_SetString(OperationalError, "can't set datestyle to ISO");
                 break;
             }
-            CLEARPGRES(pgres);
+            CLEARPGRES(self->pgres);
 
             Dprintf("conn_poll: status -> CONN_STATUS_READY");
             self->status = CONN_STATUS_READY;
@@ -1040,6 +1041,35 @@ _conn_poll_setup_async(connectionObject *self)
 }
 
 
+static cursorObject *
+_conn_get_async_cursor(connectionObject *self) {
+    PyObject *py_curs;
+
+    if (!(self->async_cursor)) {
+        PyErr_SetString(PyExc_SystemError,
+            "unexpectedly, there's no async cursor here");
+        goto error;
+    }
+
+    if (!(py_curs = PyWeakref_GetObject(self->async_cursor))) {
+        PyErr_SetString(PyExc_SystemError,
+            "got null dereferencing cursor weakref");
+        goto error;
+    }
+    if (Py_None == py_curs) {
+        PyErr_SetString(InterfaceError,
+            "the asynchronous cursor has disappeared");
+        goto error;
+    }
+
+    Py_INCREF(py_curs);
+    return (cursorObject *)py_curs;
+
+error:
+    pq_clear_async(self);
+    return NULL;
+}
+
 /* conn_poll - Main polling switch
  *
  * The function is called in all the states and connection types and invokes
@@ -1054,12 +1084,13 @@ conn_poll(connectionObject *self)
 
     switch (self->status) {
     case CONN_STATUS_SETUP:
-        Dprintf("conn_poll: status -> CONN_STATUS_CONNECTING");
+        Dprintf("conn_poll: status -> CONN_STATUS_SETUP");
         self->status = CONN_STATUS_CONNECTING;
         res = PSYCO_POLL_WRITE;
         break;
 
     case CONN_STATUS_CONNECTING:
+        Dprintf("conn_poll: status -> CONN_STATUS_CONNECTING");
         res = _conn_poll_connecting(self);
         if (res == PSYCO_POLL_OK && self->async) {
             res = _conn_poll_setup_async(self);
@@ -1067,30 +1098,28 @@ conn_poll(connectionObject *self)
         break;
 
     case CONN_STATUS_DATESTYLE:
+        Dprintf("conn_poll: status -> CONN_STATUS_DATESTYLE");
         res = _conn_poll_setup_async(self);
         break;
 
     case CONN_STATUS_READY:
     case CONN_STATUS_BEGIN:
     case CONN_STATUS_PREPARED:
+        Dprintf("conn_poll: status -> CONN_STATUS_*");
         res = _conn_poll_query(self);
 
-        if (res == PSYCO_POLL_OK && self->async && self->async_cursor) {
+        if (res == PSYCO_POLL_OK && self->async) {
+            cursorObject *curs;
+
             /* An async query has just finished: parse the tuple in the
              * target cursor. */
-            cursorObject *curs;
-            PyObject *py_curs = PyWeakref_GetObject(self->async_cursor);
-            if (Py_None == py_curs) {
-                pq_clear_async(self);
-                PyErr_SetString(InterfaceError,
-                    "the asynchronous cursor has disappeared");
+            if (!(curs = _conn_get_async_cursor(self))) {
                 res = PSYCO_POLL_ERROR;
                 break;
             }
 
-            curs = (cursorObject *)py_curs;
-            CLEARPGRES(curs->pgres);
-            curs->pgres = pq_get_last_result(self);
+            curs_set_result(curs, self->pgres);
+            self->pgres = NULL;
 
             /* fetch the tuples (if there are any) and build the result. We
              * don't care if pq_fetch return 0 or 1, but if there was an error,
@@ -1100,6 +1129,7 @@ conn_poll(connectionObject *self)
             }
 
             /* We have finished with our async_cursor */
+            Py_DECREF(curs);
             Py_CLEAR(self->async_cursor);
         }
         break;
@@ -1109,6 +1139,7 @@ conn_poll(connectionObject *self)
         res = PSYCO_POLL_ERROR;
     }
 
+    Dprintf("conn_poll: returning %d", res);
     return res;
 }
 
@@ -1189,8 +1220,6 @@ conn_set_session(connectionObject *self, int autocommit,
         int isolevel, int readonly, int deferrable)
 {
     int rv = -1;
-    PGresult *pgres = NULL;
-    char *error = NULL;
     int want_autocommit = autocommit == SRV_STATE_UNCHANGED ?
         self->autocommit : autocommit;
 
@@ -1220,21 +1249,21 @@ conn_set_session(connectionObject *self, int autocommit,
         if (isolevel != SRV_STATE_UNCHANGED) {
             if (0 > pq_set_guc_locked(self,
                     "default_transaction_isolation", srv_isolevels[isolevel],
-                    &pgres, &error, &_save)) {
+                    &_save)) {
                 goto endlock;
             }
         }
         if (readonly != SRV_STATE_UNCHANGED) {
             if (0 > pq_set_guc_locked(self,
                     "default_transaction_read_only", srv_state_guc[readonly],
-                    &pgres, &error, &_save)) {
+                    &_save)) {
                 goto endlock;
             }
         }
         if (deferrable != SRV_STATE_UNCHANGED) {
             if (0 > pq_set_guc_locked(self,
                     "default_transaction_deferrable", srv_state_guc[deferrable],
-                    &pgres, &error, &_save)) {
+                    &_save)) {
                 goto endlock;
             }
         }
@@ -1245,21 +1274,21 @@ conn_set_session(connectionObject *self, int autocommit,
         if (self->isolevel != ISOLATION_LEVEL_DEFAULT) {
             if (0 > pq_set_guc_locked(self,
                     "default_transaction_isolation", "default",
-                    &pgres, &error, &_save)) {
+                    &_save)) {
                 goto endlock;
             }
         }
         if (self->readonly != STATE_DEFAULT) {
             if (0 > pq_set_guc_locked(self,
                     "default_transaction_read_only", "default",
-                    &pgres, &error, &_save)) {
+                    &_save)) {
                 goto endlock;
             }
         }
         if (self->server_version >= 90100 && self->deferrable != STATE_DEFAULT) {
             if (0 > pq_set_guc_locked(self,
                     "default_transaction_deferrable", "default",
-                    &pgres, &error, &_save)) {
+                    &_save)) {
                 goto endlock;
             }
         }
@@ -1284,7 +1313,7 @@ endlock:
     Py_END_ALLOW_THREADS;
 
     if (rv < 0) {
-        pq_complete_error(self, &pgres, &error);
+        pq_complete_error(self);
         goto exit;
     }
 
@@ -1303,8 +1332,6 @@ exit:
 RAISES_NEG int
 conn_set_client_encoding(connectionObject *self, const char *pgenc)
 {
-    PGresult *pgres = NULL;
-    char *error = NULL;
     int res = -1;
     char *clean_enc = NULL;
 
@@ -1320,12 +1347,11 @@ conn_set_client_encoding(connectionObject *self, const char *pgenc)
 
     /* abort the current transaction, to set the encoding ouside of
        transactions */
-    if ((res = pq_abort_locked(self, &pgres, &error, &_save))) {
+    if ((res = pq_abort_locked(self, &_save))) {
         goto endlock;
     }
 
-    if ((res = pq_set_guc_locked(self, "client_encoding", clean_enc,
-            &pgres, &error, &_save))) {
+    if ((res = pq_set_guc_locked(self, "client_encoding", clean_enc, &_save))) {
         goto endlock;
     }
 
@@ -1334,7 +1360,7 @@ endlock:
     Py_END_ALLOW_THREADS;
 
     if (res < 0) {
-        pq_complete_error(self, &pgres, &error);
+        pq_complete_error(self);
         goto exit;
     }
 
@@ -1360,18 +1386,15 @@ exit:
 RAISES_NEG int
 conn_tpc_begin(connectionObject *self, xidObject *xid)
 {
-    PGresult *pgres = NULL;
-    char *error = NULL;
-
     Dprintf("conn_tpc_begin: starting transaction");
 
     Py_BEGIN_ALLOW_THREADS;
     pthread_mutex_lock(&self->lock);
 
-    if (pq_begin_locked(self, &pgres, &error, &_save) < 0) {
+    if (pq_begin_locked(self, &_save) < 0) {
         pthread_mutex_unlock(&(self->lock));
         Py_BLOCK_THREADS;
-        pq_complete_error(self, &pgres, &error);
+        pq_complete_error(self);
         return -1;
     }
 
@@ -1394,8 +1417,6 @@ conn_tpc_begin(connectionObject *self, xidObject *xid)
 RAISES_NEG int
 conn_tpc_command(connectionObject *self, const char *cmd, xidObject *xid)
 {
-    PGresult *pgres = NULL;
-    char *error = NULL;
     PyObject *tid = NULL;
     const char *ctid;
     int rv = -1;
@@ -1403,17 +1424,16 @@ conn_tpc_command(connectionObject *self, const char *cmd, xidObject *xid)
     Dprintf("conn_tpc_command: %s", cmd);
 
     /* convert the xid into PostgreSQL transaction id while keeping the GIL */
-    if (!(tid = psycopg_ensure_bytes(xid_get_tid(xid)))) { goto exit; }
+    if (!(tid = psyco_ensure_bytes(xid_get_tid(xid)))) { goto exit; }
     if (!(ctid = Bytes_AsString(tid))) { goto exit; }
 
     Py_BEGIN_ALLOW_THREADS;
     pthread_mutex_lock(&self->lock);
 
-    if (0 > (rv = pq_tpc_command_locked(self, cmd, ctid,
-                                        &pgres, &error, &_save))) {
+    if (0 > (rv = pq_tpc_command_locked(self, cmd, ctid, &_save))) {
         pthread_mutex_unlock(&self->lock);
         Py_BLOCK_THREADS;
-        pq_complete_error(self, &pgres, &error);
+        pq_complete_error(self);
         goto exit;
     }
 
@@ -1458,3 +1478,24 @@ exit:
     return rv;
 
 }
+
+
+void
+conn_set_result(connectionObject *self, PGresult *pgres)
+{
+    PQclear(self->pgres);
+    self->pgres = pgres;
+}
+
+
+void
+conn_set_error(connectionObject *self, const char *msg)
+{
+    if (self->error) {
+        free(self->error);
+        self->error = NULL;
+    }
+    if (msg && *msg) {
+        self->error = strdup(msg);
+    }
+}
