@@ -2,7 +2,7 @@
 #
 # extras_dictcursor - test if DictCursor extension class works
 #
-# Copyright (C) 2004-2010 Federico Di Gregorio  <fog@debian.org>
+# Copyright (C) 2004-2019 Federico Di Gregorio  <fog@debian.org>
 #
 # psycopg2 is free software: you can redistribute it and/or modify it
 # under the terms of the GNU Lesser General Public License as published
@@ -15,16 +15,20 @@
 # License for more details.
 
 import time
+import pickle
+import unittest
 from datetime import timedelta
+
 import psycopg2
+from psycopg2.compat import lru_cache
 import psycopg2.extras
-from testutils import unittest, ConnectingTestCase, skip_before_postgres
-from testutils import skip_if_no_namedtuple
+from psycopg2.extras import NamedTupleConnection, NamedTupleCursor
+
+from .testutils import ConnectingTestCase, skip_before_postgres, \
+    skip_before_python, skip_from_python
 
 
-class ExtrasDictCursorTests(ConnectingTestCase):
-    """Test if DictCursor extension class works."""
-
+class _DictCursorBase(ConnectingTestCase):
     def setUp(self):
         ConnectingTestCase.setUp(self)
         curs = self.conn.cursor()
@@ -32,6 +36,30 @@ class ExtrasDictCursorTests(ConnectingTestCase):
         curs.execute("INSERT INTO ExtrasDictCursorTests VALUES ('bar')")
         self.conn.commit()
 
+    def _testIterRowNumber(self, curs):
+        # Only checking for dataset < itersize:
+        # see CursorTests.test_iter_named_cursor_rownumber
+        curs.itersize = 20
+        curs.execute("""select * from generate_series(1,10)""")
+        for i, r in enumerate(curs):
+            self.assertEqual(i + 1, curs.rownumber)
+
+    def _testNamedCursorNotGreedy(self, curs):
+        curs.itersize = 2
+        curs.execute("""select clock_timestamp() as ts from generate_series(1,3)""")
+        recs = []
+        for t in curs:
+            time.sleep(0.01)
+            recs.append(t)
+
+        # check that the dataset was not fetched in a single gulp
+        self.assert_(recs[1]['ts'] - recs[0]['ts'] < timedelta(seconds=0.005))
+        self.assert_(recs[2]['ts'] - recs[1]['ts'] > timedelta(seconds=0.0099))
+
+
+class ExtrasDictCursorTests(_DictCursorBase):
+    """Test if DictCursor extension class works."""
+
     def testDictConnCursorArgs(self):
         self.conn.close()
         self.conn = self.connect(connection_factory=psycopg2.extras.DictConnection)
@@ -81,35 +109,6 @@ class ExtrasDictCursorTests(ConnectingTestCase):
         self.failUnless(row[0] == 'bar')
         return row
 
-    def testDictCursorWithPlainCursorRealFetchOne(self):
-        self._testWithPlainCursorReal(lambda curs: curs.fetchone())
-
-    def testDictCursorWithPlainCursorRealFetchMany(self):
-        self._testWithPlainCursorReal(lambda curs: curs.fetchmany(100)[0])
-
-    def testDictCursorWithPlainCursorRealFetchManyNoarg(self):
-        self._testWithPlainCursorReal(lambda curs: curs.fetchmany()[0])
-
-    def testDictCursorWithPlainCursorRealFetchAll(self):
-        self._testWithPlainCursorReal(lambda curs: curs.fetchall()[0])
-
-    def testDictCursorWithPlainCursorRealIter(self):
-        def getter(curs):
-            for row in curs:
-                return row
-        self._testWithPlainCursorReal(getter)
-
-    @skip_before_postgres(8, 0)
-    def testDictCursorWithPlainCursorRealIterRowNumber(self):
-        curs = self.conn.cursor(cursor_factory=psycopg2.extras.RealDictCursor)
-        self._testIterRowNumber(curs)
-
-    def _testWithPlainCursorReal(self, getter):
-        curs = self.conn.cursor(cursor_factory=psycopg2.extras.RealDictCursor)
-        curs.execute("SELECT * FROM ExtrasDictCursorTests")
-        row = getter(curs)
-        self.failUnless(row['foo'] == 'bar')
-
     def testDictCursorWithNamedCursorFetchOne(self):
         self._testWithNamedCursor(lambda curs: curs.fetchone())
 
@@ -145,6 +144,123 @@ class ExtrasDictCursorTests(ConnectingTestCase):
         self.failUnless(row['foo'] == 'bar')
         self.failUnless(row[0] == 'bar')
 
+    def testPickleDictRow(self):
+        curs = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
+        curs.execute("select 10 as a, 20 as b")
+        r = curs.fetchone()
+        d = pickle.dumps(r)
+        r1 = pickle.loads(d)
+        self.assertEqual(r, r1)
+        self.assertEqual(r[0], r1[0])
+        self.assertEqual(r[1], r1[1])
+        self.assertEqual(r['a'], r1['a'])
+        self.assertEqual(r['b'], r1['b'])
+        self.assertEqual(r._index, r1._index)
+
+    @skip_from_python(3)
+    def test_iter_methods_2(self):
+        curs = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
+        curs.execute("select 10 as a, 20 as b")
+        r = curs.fetchone()
+        self.assert_(isinstance(r.keys(), list))
+        self.assertEqual(len(r.keys()), 2)
+        self.assert_(isinstance(r.values(), tuple))     # sic?
+        self.assertEqual(len(r.values()), 2)
+        self.assert_(isinstance(r.items(), list))
+        self.assertEqual(len(r.items()), 2)
+
+        self.assert_(not isinstance(r.iterkeys(), list))
+        self.assertEqual(len(list(r.iterkeys())), 2)
+        self.assert_(not isinstance(r.itervalues(), list))
+        self.assertEqual(len(list(r.itervalues())), 2)
+        self.assert_(not isinstance(r.iteritems(), list))
+        self.assertEqual(len(list(r.iteritems())), 2)
+
+    @skip_before_python(3)
+    def test_iter_methods_3(self):
+        curs = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
+        curs.execute("select 10 as a, 20 as b")
+        r = curs.fetchone()
+        self.assert_(not isinstance(r.keys(), list))
+        self.assertEqual(len(list(r.keys())), 2)
+        self.assert_(not isinstance(r.values(), list))
+        self.assertEqual(len(list(r.values())), 2)
+        self.assert_(not isinstance(r.items(), list))
+        self.assertEqual(len(list(r.items())), 2)
+
+    def test_order(self):
+        curs = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
+        curs.execute("select 5 as foo, 4 as bar, 33 as baz, 2 as qux")
+        r = curs.fetchone()
+        self.assertEqual(list(r), [5, 4, 33, 2])
+        self.assertEqual(list(r.keys()), ['foo', 'bar', 'baz', 'qux'])
+        self.assertEqual(list(r.values()), [5, 4, 33, 2])
+        self.assertEqual(list(r.items()),
+            [('foo', 5), ('bar', 4), ('baz', 33), ('qux', 2)])
+
+        r1 = pickle.loads(pickle.dumps(r))
+        self.assertEqual(list(r1), list(r))
+        self.assertEqual(list(r1.keys()), list(r.keys()))
+        self.assertEqual(list(r1.values()), list(r.values()))
+        self.assertEqual(list(r1.items()), list(r.items()))
+
+    @skip_from_python(3)
+    def test_order_iter(self):
+        curs = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
+        curs.execute("select 5 as foo, 4 as bar, 33 as baz, 2 as qux")
+        r = curs.fetchone()
+        self.assertEqual(list(r.iterkeys()), ['foo', 'bar', 'baz', 'qux'])
+        self.assertEqual(list(r.itervalues()), [5, 4, 33, 2])
+        self.assertEqual(list(r.iteritems()),
+            [('foo', 5), ('bar', 4), ('baz', 33), ('qux', 2)])
+
+        r1 = pickle.loads(pickle.dumps(r))
+        self.assertEqual(list(r1.iterkeys()), list(r.iterkeys()))
+        self.assertEqual(list(r1.itervalues()), list(r.itervalues()))
+        self.assertEqual(list(r1.iteritems()), list(r.iteritems()))
+
+
+class ExtrasDictCursorRealTests(_DictCursorBase):
+    def testDictCursorWithPlainCursorRealFetchOne(self):
+        self._testWithPlainCursorReal(lambda curs: curs.fetchone())
+
+    def testDictCursorWithPlainCursorRealFetchMany(self):
+        self._testWithPlainCursorReal(lambda curs: curs.fetchmany(100)[0])
+
+    def testDictCursorWithPlainCursorRealFetchManyNoarg(self):
+        self._testWithPlainCursorReal(lambda curs: curs.fetchmany()[0])
+
+    def testDictCursorWithPlainCursorRealFetchAll(self):
+        self._testWithPlainCursorReal(lambda curs: curs.fetchall()[0])
+
+    def testDictCursorWithPlainCursorRealIter(self):
+        def getter(curs):
+            for row in curs:
+                return row
+        self._testWithPlainCursorReal(getter)
+
+    @skip_before_postgres(8, 0)
+    def testDictCursorWithPlainCursorRealIterRowNumber(self):
+        curs = self.conn.cursor(cursor_factory=psycopg2.extras.RealDictCursor)
+        self._testIterRowNumber(curs)
+
+    def _testWithPlainCursorReal(self, getter):
+        curs = self.conn.cursor(cursor_factory=psycopg2.extras.RealDictCursor)
+        curs.execute("SELECT * FROM ExtrasDictCursorTests")
+        row = getter(curs)
+        self.failUnless(row['foo'] == 'bar')
+
+    def testPickleRealDictRow(self):
+        curs = self.conn.cursor(cursor_factory=psycopg2.extras.RealDictCursor)
+        curs.execute("select 10 as a, 20 as b")
+        r = curs.fetchone()
+        d = pickle.dumps(r)
+        r1 = pickle.loads(d)
+        self.assertEqual(r, r1)
+        self.assertEqual(r['a'], r1['a'])
+        self.assertEqual(r['b'], r1['b'])
+        self.assertEqual(r._column_mapping, r1._column_mapping)
+
     def testDictCursorRealWithNamedCursorFetchOne(self):
         self._testWithNamedCursorReal(lambda curs: curs.fetchone())
 
@@ -180,62 +296,72 @@ class ExtrasDictCursorTests(ConnectingTestCase):
         row = getter(curs)
         self.failUnless(row['foo'] == 'bar')
 
-    def _testNamedCursorNotGreedy(self, curs):
-        curs.itersize = 2
-        curs.execute("""select clock_timestamp() as ts from generate_series(1,3)""")
-        recs = []
-        for t in curs:
-            time.sleep(0.01)
-            recs.append(t)
-
-        # check that the dataset was not fetched in a single gulp
-        self.assert_(recs[1]['ts'] - recs[0]['ts'] < timedelta(seconds=0.005))
-        self.assert_(recs[2]['ts'] - recs[1]['ts'] > timedelta(seconds=0.0099))
-
-    def _testIterRowNumber(self, curs):
-        # Only checking for dataset < itersize:
-        # see CursorTests.test_iter_named_cursor_rownumber
-        curs.itersize = 20
-        curs.execute("""select * from generate_series(1,10)""")
-        for i, r in enumerate(curs):
-            self.assertEqual(i + 1, curs.rownumber)
-
-    def testPickleDictRow(self):
-        import pickle
-        curs = self.conn.cursor(cursor_factory=psycopg2.extras.DictCursor)
-        curs.execute("select 10 as a, 20 as b")
-        r = curs.fetchone()
-        d = pickle.dumps(r)
-        r1 = pickle.loads(d)
-        self.assertEqual(r, r1)
-        self.assertEqual(r[0], r1[0])
-        self.assertEqual(r[1], r1[1])
-        self.assertEqual(r['a'], r1['a'])
-        self.assertEqual(r['b'], r1['b'])
-        self.assertEqual(r._index, r1._index)
-
-    def testPickleRealDictRow(self):
-        import pickle
+    @skip_from_python(3)
+    def test_iter_methods_2(self):
         curs = self.conn.cursor(cursor_factory=psycopg2.extras.RealDictCursor)
         curs.execute("select 10 as a, 20 as b")
         r = curs.fetchone()
-        d = pickle.dumps(r)
-        r1 = pickle.loads(d)
-        self.assertEqual(r, r1)
-        self.assertEqual(r['a'], r1['a'])
-        self.assertEqual(r['b'], r1['b'])
-        self.assertEqual(r._column_mapping, r1._column_mapping)
+        self.assert_(isinstance(r.keys(), list))
+        self.assertEqual(len(r.keys()), 2)
+        self.assert_(isinstance(r.values(), list))
+        self.assertEqual(len(r.values()), 2)
+        self.assert_(isinstance(r.items(), list))
+        self.assertEqual(len(r.items()), 2)
+
+        self.assert_(not isinstance(r.iterkeys(), list))
+        self.assertEqual(len(list(r.iterkeys())), 2)
+        self.assert_(not isinstance(r.itervalues(), list))
+        self.assertEqual(len(list(r.itervalues())), 2)
+        self.assert_(not isinstance(r.iteritems(), list))
+        self.assertEqual(len(list(r.iteritems())), 2)
+
+    @skip_before_python(3)
+    def test_iter_methods_3(self):
+        curs = self.conn.cursor(cursor_factory=psycopg2.extras.RealDictCursor)
+        curs.execute("select 10 as a, 20 as b")
+        r = curs.fetchone()
+        self.assert_(not isinstance(r.keys(), list))
+        self.assertEqual(len(list(r.keys())), 2)
+        self.assert_(not isinstance(r.values(), list))
+        self.assertEqual(len(list(r.values())), 2)
+        self.assert_(not isinstance(r.items(), list))
+        self.assertEqual(len(list(r.items())), 2)
+
+    def test_order(self):
+        curs = self.conn.cursor(cursor_factory=psycopg2.extras.RealDictCursor)
+        curs.execute("select 5 as foo, 4 as bar, 33 as baz, 2 as qux")
+        r = curs.fetchone()
+        self.assertEqual(list(r), ['foo', 'bar', 'baz', 'qux'])
+        self.assertEqual(list(r.keys()), ['foo', 'bar', 'baz', 'qux'])
+        self.assertEqual(list(r.values()), [5, 4, 33, 2])
+        self.assertEqual(list(r.items()),
+            [('foo', 5), ('bar', 4), ('baz', 33), ('qux', 2)])
+
+        r1 = pickle.loads(pickle.dumps(r))
+        self.assertEqual(list(r1), list(r))
+        self.assertEqual(list(r1.keys()), list(r.keys()))
+        self.assertEqual(list(r1.values()), list(r.values()))
+        self.assertEqual(list(r1.items()), list(r.items()))
+
+    @skip_from_python(3)
+    def test_order_iter(self):
+        curs = self.conn.cursor(cursor_factory=psycopg2.extras.RealDictCursor)
+        curs.execute("select 5 as foo, 4 as bar, 33 as baz, 2 as qux")
+        r = curs.fetchone()
+        self.assertEqual(list(r.iterkeys()), ['foo', 'bar', 'baz', 'qux'])
+        self.assertEqual(list(r.itervalues()), [5, 4, 33, 2])
+        self.assertEqual(list(r.iteritems()),
+            [('foo', 5), ('bar', 4), ('baz', 33), ('qux', 2)])
+
+        r1 = pickle.loads(pickle.dumps(r))
+        self.assertEqual(list(r1.iterkeys()), list(r.iterkeys()))
+        self.assertEqual(list(r1.itervalues()), list(r.itervalues()))
+        self.assertEqual(list(r1.iteritems()), list(r.iteritems()))
 
 
 class NamedTupleCursorTest(ConnectingTestCase):
     def setUp(self):
         ConnectingTestCase.setUp(self)
-        from psycopg2.extras import NamedTupleConnection
-
-        try:
-            from collections import namedtuple      # noqa
-        except ImportError:
-            return
 
         self.conn = self.connect(connection_factory=NamedTupleConnection)
         curs = self.conn.cursor()
@@ -245,13 +371,11 @@ class NamedTupleCursorTest(ConnectingTestCase):
         curs.execute("INSERT INTO nttest VALUES (3, 'baz')")
         self.conn.commit()
 
-    @skip_if_no_namedtuple
     def test_cursor_args(self):
         cur = self.conn.cursor('foo', cursor_factory=psycopg2.extras.DictCursor)
         self.assertEqual(cur.name, 'foo')
         self.assert_(isinstance(cur, psycopg2.extras.DictCursor))
 
-    @skip_if_no_namedtuple
     def test_fetchone(self):
         curs = self.conn.cursor()
         curs.execute("select * from nttest order by 1")
@@ -263,7 +387,6 @@ class NamedTupleCursorTest(ConnectingTestCase):
         self.assertEqual(curs.rownumber, 1)
         self.assertEqual(curs.rowcount, 3)
 
-    @skip_if_no_namedtuple
     def test_fetchmany_noarg(self):
         curs = self.conn.cursor()
         curs.arraysize = 2
@@ -277,7 +400,6 @@ class NamedTupleCursorTest(ConnectingTestCase):
         self.assertEqual(curs.rownumber, 2)
         self.assertEqual(curs.rowcount, 3)
 
-    @skip_if_no_namedtuple
     def test_fetchmany(self):
         curs = self.conn.cursor()
         curs.execute("select * from nttest order by 1")
@@ -290,7 +412,6 @@ class NamedTupleCursorTest(ConnectingTestCase):
         self.assertEqual(curs.rownumber, 2)
         self.assertEqual(curs.rowcount, 3)
 
-    @skip_if_no_namedtuple
     def test_fetchall(self):
         curs = self.conn.cursor()
         curs.execute("select * from nttest order by 1")
@@ -305,7 +426,6 @@ class NamedTupleCursorTest(ConnectingTestCase):
         self.assertEqual(curs.rownumber, 3)
         self.assertEqual(curs.rowcount, 3)
 
-    @skip_if_no_namedtuple
     def test_executemany(self):
         curs = self.conn.cursor()
         curs.executemany("delete from nttest where i = %s",
@@ -316,52 +436,31 @@ class NamedTupleCursorTest(ConnectingTestCase):
         self.assertEqual(res[0].i, 3)
         self.assertEqual(res[0].s, 'baz')
 
-    @skip_if_no_namedtuple
     def test_iter(self):
         curs = self.conn.cursor()
         curs.execute("select * from nttest order by 1")
         i = iter(curs)
         self.assertEqual(curs.rownumber, 0)
 
-        t = i.next()
+        t = next(i)
         self.assertEqual(t.i, 1)
         self.assertEqual(t.s, 'foo')
         self.assertEqual(curs.rownumber, 1)
         self.assertEqual(curs.rowcount, 3)
 
-        t = i.next()
+        t = next(i)
         self.assertEqual(t.i, 2)
         self.assertEqual(t.s, 'bar')
         self.assertEqual(curs.rownumber, 2)
         self.assertEqual(curs.rowcount, 3)
 
-        t = i.next()
+        t = next(i)
         self.assertEqual(t.i, 3)
         self.assertEqual(t.s, 'baz')
-        self.assertRaises(StopIteration, i.next)
+        self.assertRaises(StopIteration, next, i)
         self.assertEqual(curs.rownumber, 3)
         self.assertEqual(curs.rowcount, 3)
 
-    def test_error_message(self):
-        try:
-            from collections import namedtuple          # noqa
-        except ImportError:
-            # an import error somewhere
-            from psycopg2.extras import NamedTupleConnection
-            try:
-                self.conn = self.connect(
-                    connection_factory=NamedTupleConnection)
-                curs = self.conn.cursor()
-                curs.execute("select 1")
-                curs.fetchone()
-            except ImportError:
-                pass
-            else:
-                self.fail("expecting ImportError")
-        else:
-            return self.skipTest("namedtuple available")
-
-    @skip_if_no_namedtuple
     def test_record_updated(self):
         curs = self.conn.cursor()
         curs.execute("select 1 as foo;")
@@ -373,7 +472,6 @@ class NamedTupleCursorTest(ConnectingTestCase):
         self.assertEqual(r.bar, 2)
         self.assertRaises(AttributeError, getattr, r, 'foo')
 
-    @skip_if_no_namedtuple
     def test_no_result_no_surprise(self):
         curs = self.conn.cursor()
         curs.execute("update nttest set s = s")
@@ -382,7 +480,6 @@ class NamedTupleCursorTest(ConnectingTestCase):
         curs.execute("update nttest set s = s")
         self.assertRaises(psycopg2.ProgrammingError, curs.fetchall)
 
-    @skip_if_no_namedtuple
     def test_bad_col_names(self):
         curs = self.conn.cursor()
         curs.execute('select 1 as "foo.bar_baz", 2 as "?column?", 3 as "3"')
@@ -391,10 +488,16 @@ class NamedTupleCursorTest(ConnectingTestCase):
         self.assertEqual(rv.f_column_, 2)
         self.assertEqual(rv.f3, 3)
 
-    @skip_if_no_namedtuple
+    @skip_before_python(3)
+    @skip_before_postgres(8)
+    def test_nonascii_name(self):
+        curs = self.conn.cursor()
+        curs.execute('select 1 as \xe5h\xe9')
+        rv = curs.fetchone()
+        self.assertEqual(getattr(rv, '\xe5h\xe9'), 1)
+
     def test_minimal_generation(self):
         # Instrument the class to verify it gets called the minimum number of times.
-        from psycopg2.extras import NamedTupleCursor
         f_orig = NamedTupleCursor._make_nt
         calls = [0]
 
@@ -425,7 +528,6 @@ class NamedTupleCursorTest(ConnectingTestCase):
         finally:
             NamedTupleCursor._make_nt = f_orig
 
-    @skip_if_no_namedtuple
     @skip_before_postgres(8, 0)
     def test_named(self):
         curs = self.conn.cursor('tmp')
@@ -434,30 +536,26 @@ class NamedTupleCursorTest(ConnectingTestCase):
         recs.extend(curs.fetchmany(5))
         recs.append(curs.fetchone())
         recs.extend(curs.fetchall())
-        self.assertEqual(range(10), [t.i for t in recs])
+        self.assertEqual(list(range(10)), [t.i for t in recs])
 
-    @skip_if_no_namedtuple
     def test_named_fetchone(self):
         curs = self.conn.cursor('tmp')
         curs.execute("""select 42 as i""")
         t = curs.fetchone()
         self.assertEqual(t.i, 42)
 
-    @skip_if_no_namedtuple
     def test_named_fetchmany(self):
         curs = self.conn.cursor('tmp')
         curs.execute("""select 42 as i""")
         recs = curs.fetchmany(10)
         self.assertEqual(recs[0].i, 42)
 
-    @skip_if_no_namedtuple
     def test_named_fetchall(self):
         curs = self.conn.cursor('tmp')
         curs.execute("""select 42 as i""")
         recs = curs.fetchall()
         self.assertEqual(recs[0].i, 42)
 
-    @skip_if_no_namedtuple
     @skip_before_postgres(8, 2)
     def test_not_greedy(self):
         curs = self.conn.cursor('tmp')
@@ -472,7 +570,6 @@ class NamedTupleCursorTest(ConnectingTestCase):
         self.assert_(recs[1].ts - recs[0].ts < timedelta(seconds=0.005))
         self.assert_(recs[2].ts - recs[1].ts > timedelta(seconds=0.0099))
 
-    @skip_if_no_namedtuple
     @skip_before_postgres(8, 0)
     def test_named_rownumber(self):
         curs = self.conn.cursor('tmp')
@@ -483,9 +580,46 @@ class NamedTupleCursorTest(ConnectingTestCase):
         for i, t in enumerate(curs):
             self.assertEqual(i + 1, curs.rownumber)
 
+    def test_cache(self):
+        curs = self.conn.cursor()
+        curs.execute("select 10 as a, 20 as b")
+        r1 = curs.fetchone()
+        curs.execute("select 10 as a, 20 as c")
+        r2 = curs.fetchone()
+        curs.execute("select 10 as a, 30 as b")
+        r3 = curs.fetchone()
+
+        self.assert_(type(r1) is type(r3))
+        self.assert_(type(r1) is not type(r2))
+
+    def test_max_cache(self):
+        old_func = NamedTupleCursor._cached_make_nt
+        NamedTupleCursor._cached_make_nt = \
+            lru_cache(8)(NamedTupleCursor._do_make_nt)
+        try:
+            recs = []
+            curs = self.conn.cursor()
+            for i in range(10):
+                curs.execute("select 1 as f%s" % i)
+                recs.append(curs.fetchone())
+
+            # Still in cache
+            curs.execute("select 1 as f9")
+            rec = curs.fetchone()
+            self.assert_(any(type(r) is type(rec) for r in recs))
+
+            # Gone from cache
+            curs.execute("select 1 as f0")
+            rec = curs.fetchone()
+            self.assert_(all(type(r) is not type(rec) for r in recs))
+
+        finally:
+            NamedTupleCursor._cached_make_nt = old_func
+
 
 def test_suite():
     return unittest.TestLoader().loadTestsFromName(__name__)
 
+
 if __name__ == "__main__":
     unittest.main()
